import argparse
import itertools
import json
import logging
import subprocess
import sys
import time
import webbrowser

import colored
import requests
import yaml

from atils import atils_kubernetes as k8s_utils
from atils.common import config, template_utils
from atils.common.settings import settings
from kubernetes import client
from kubernetes import config as k8s_config

client.rest.logger.setLevel(logging.WARNING)

logging.basicConfig(level=config.get_logging_level())  # type: ignore


def main(args: list[str]) -> None:
    # This variable tracks whether or not we have configuration available to run kubernetes commands
    CAN_RUN: bool = k8s_utils.load_config()

    if not CAN_RUN:
        logging.error("No configuration available to run kubernetes commands")
        exit(1)

    # TODO Add auto-complete for application names
    # TODO Add a command to list all apps and their status
    parser: argparse.ArgumentParser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        help="Commands to manage argocd", dest="subparser_name"
    )

    install_parser = subparsers.add_parser("install", help="Install and set up ArgoCD")
    install_parser.add_argument(
        "environment",
        choices=["dev-laptop", "dev-cluster", "prod-cluster"],
        help="Which environment to use for templating",
    )

    port_forward_parser = subparsers.add_parser(
        "port-forward",
        help="Automatically port fowards ArgoCD, opens a browser, and copies the password",
    )

    grab_password_parser = subparsers.add_parser(
        "get-password", help="Get the admin password using the default secret."
    )

    print_apps_parser = subparsers.add_parser("test")

    disable_parser = subparsers.add_parser(
        "disable",
        help="Disables an application from the master stack",
    )
    disable_parser.add_argument("application", help="The application to disable")

    enable_parser = subparsers.add_parser(
        "enable",
        help="Ensables an application from the master stack",
    )
    enable_parser.add_argument("application", help="The application to enable")

    if len(args) == 0:
        parser.print_help(sys.stderr)
        sys.exit(1)

    arguments: argparse.Namespace = parser.parse_args(args)

    # Install ArgoCD using Helm
    if arguments.subparser_name == "install":
        args_dict = vars(arguments)
        # Check if we provided an environment argument, and install it if not
        if args_dict["environment"] is None:
            logging.error("Please provide an environment")
            sys.exit(1)
        else:
            setup_argocd(args_dict["environment"])
    # Port forward ArgoCD and open a browser window
    elif arguments.subparser_name == "port-forward":
        open_argocd_port_forward()
    # Disable an application from the master stack
    elif arguments.subparser_name == "disable":
        args_dict = vars(arguments)
        # Check if we provided an application argument, and disable it if we did
        if args_dict["application"] is None:
            logging.error("Please provide an application name")
            sys.exit(1)
        else:
            disable_application(args_dict["application"])
    # Enable an application from the master stack
    elif arguments.subparser_name == "enable":
        args_dict = vars(arguments)
        # Check if we provided an application argument, and enable it if we did
        if args_dict["application"] is None:
            logging.error("Please provide an application name")
            sys.exit(1)
        else:
            enable_application(args_dict["application"])
    # Get the ArgoCD password from the default secret
    # TODO once we've validated that ArgoCD doesn't do its own auth, delete this
    elif arguments.subparser_name == "get-password":
        get_argocd_password()
    # TODO Figure out what the hell was going on here
    elif arguments.subparser_name == "test":
        print_application_status(["grafana-loki", "gateways"], True)
    else:
        logging.error(f"Invalid command: ${arguments.subparser_name}")
        exit(1)


def get_argocd_password() -> None:
    # TODO once we've validated that ArgoCD doesn't do its own auth, delete this
    # Use kubectl to get the ArgoCD password
    result = subprocess.run(
        "kubectl -n argocd get secret argocd-initial-admin-secret -o jsonpath={{.data.password}} | base64 -d".format(),
        shell=True,
        capture_output=True,
    )
    subprocess.run(
        f"echo {result.stdout.decode('utf-8')} | pbcopy",
        shell=True,
        capture_output=True,
    )


# TODO determine if this is going to be an internal-only function, or if we want to expost it
def sync_master_stack_application(application: str) -> None:
    """
    Run a sync operation against the master stack application, only syncing the specified application
    application

    Args:
        application (str): The name of the application to sync. It must be a sub-application of the master-stack
    """
    token = _get_argocd_bearer_token()

    response = requests.post(
        f"{settings.argocd_url}/api/v1/applications/master-stack/sync",
        data=json.dumps(
            {
                "resources": [
                    {
                        "kind": "Application",
                        "name": application,
                        "namespace": "argocd",
                        "group": "argoproj.io",
                    }
                ]
            }
        ),
        verify=False,
        headers={"Authorization": f"Bearer {token}"},
    )

    if response.status_code == 200:
        logging.info(f"Synced {application} successfully")
    else:
        logging.error(f"Failed to sync {application}")
        logging.error(response.text)


def enable_application(application: str) -> None:
    """
    Enable the specified application. It must be a sub-application of the master stack
    Args:
        application (str): The name of the application to enable
    """
    try:
        api_instance = client.CustomObjectsApi()

        # Get the master-stack application, located in the ArgoCD namespace. This is what should be creating all of
        # our applications, so we want disable the target application in master-stacks parameters, so it doesn't
        # recreate it
        api_response: dict = _get_master_stack_application()

        resources = api_response.get("status").get("resources")  # type: ignore

        is_active = _is_application_active(application, resources)
        is_enabled = _is_application_enabled_in_master_stack(application, api_response)

        # Constants we'll use for modifying the master-stack later
        namespace = "argocd"
        name = "master-stack"
        group = "argoproj.io"
        version = "v1alpha1"
        plural = "applications"

        if is_enabled:
            logging.info(f"Application {application} already enabled")
            exit(0)
        else:
            # Get Helm parameters for application
            source_helm_params = _get_source_helm_params(api_response)

            edited_parameters = source_helm_params

            for i in range(len(edited_parameters)):
                if application == edited_parameters[i].get("name").split(".")[0]:  # type: ignore
                    del edited_parameters[i]

            # Create JSON merge patch to update the helm parameters
            json_patch = {
                "spec": {"source": {"helm": {"parameters": edited_parameters}}}
            }

            # Use patch_namespaced_custom_object to update the application with the new helm parameter
            api_response = api_instance.patch_namespaced_custom_object(
                group, version, namespace, plural, name, json_patch
            )

            logging.info(f"Application {application} enabled successfully")
            if not is_active:
                logging.debug(f"Syncing {application} so it actually shows up")
                sync_master_stack_application(application)
            else:
                logging.debug(f"No need to sync {application}")
    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


# Sync so that the app deletes when we run this
def disable_application(application: str) -> None:
    """
    Disable the specified application. It must be a sub-application of the master stack
    Args:
        application (str): The name of the application to disable
    """
    try:
        api_instance = client.CustomObjectsApi()
        api_response = _get_master_stack_application()

        resources = api_response.get("status").get("resources")  # type: ignore

        is_active = _is_application_active(application, resources)
        is_enabled = _is_application_enabled_in_master_stack(application, api_response)

        # Constants we'll use for modifying the master-stack later
        namespace = "argocd"
        name = "master-stack"
        group = "argoproj.io"
        version = "v1alpha1"
        plural = "applications"

        if not is_enabled:
            logging.info(f"Application {application} already disabled")
            exit(0)
        else:
            # Check Helm parameters for application
            source_helm_params = _get_source_helm_params(api_response)

            # Create a new helm parameter with the given application name and value
            new_helm_param = {"name": f"{application}.enabled", "value": "false"}
            edited_parameters = source_helm_params + [new_helm_param]

            # Create JSON merge patch to update the helm parameters
            json_patch = {
                "spec": {"source": {"helm": {"parameters": edited_parameters}}}
            }

            # Use patch_namespaced_custom_object to update the application with the new helm parameter
            api_response = api_instance.patch_namespaced_custom_object(
                group, version, namespace, plural, name, json_patch
            )

            logging.info(f"{application} disabled in master-stack")

            if is_active:
                api = client.CustomObjectsApi()

                try:
                    api.delete_namespaced_custom_object(
                        group="argoproj.io",
                        version="v1alpha1",
                        namespace="argocd",
                        plural="applications",
                        name=application,
                    )
                except Exception as e:
                    logging.error(
                        f"Unable to delete application {application}, although disabling was successful"
                    )
                    exit(1)

    except Exception as e:
        print(f"Error: {e}")
        sys.exit(1)


# TODO Just read the name of our current context to get the environment
def setup_argocd(environment: str) -> None:
    """
    Create the ArgoCD namespace if it doesn't exist, install ArgoCD with Helm (we don't check if its installed
    currently), then wait for the ArgoCD pods to be ready, then install the master stack
    Args:
        environment (str): The name of the environment we want to install into. For now, the valid options are
        dev-laptop, dev-cluster, and prod-cluster. TODO Update this documentation to reflect our new naming
        convention:
        qa-cluster: The K3s cluster running on our laptop
        preprod-cluster: The RKE cluster made out of VMs we use for testing in a working environment
        prod-cluster: The RKE cluster made out of VMs used to run our actual services

    """
    # Check if the argocd namespace exists, and create it if it doesn't
    if not k8s_utils.check_namespace_exists("argocd"):
        api = client.CoreV1Api()

        # Create the namespace
        namespace_body = client.V1Namespace(metadata=client.V1ObjectMeta(name="argocd"))
        api.create_namespace(namespace_body)

    # Using the helm CLI as a dependency to add the argo repo, and install it.
    # TODO check if ArgoCD is already installed
    result = subprocess.run(
        f"""helm repo add argo https://argoproj.github.io/argo-helm && helm -n argocd upgrade --install argocd \
            -f {settings.SCRIPT_INSTALL_DIRECTORY}/../kubernetes/argocd/values.yaml argo/argo-cd""",
        shell=True,
        capture_output=True,
    )

    if result.returncode == 0:
        logging.info("ArgoCD Helm chart successfully installed")
    else:
        logging.warning(result.stdout)
        logging.warning(
            "ArgoCD Helm chart failed to install. This may simply be because it is already installed,"
            + " which we don't check"
        )

    _wait_for_argocd_pods_to_be_ready()

    _install_master_stack_if_not_present(environment)

    # TODO Get this part working, basically we want to sit until all our applications are done, or enough time passes
    # applications = get_tracked_applications()
    # LINE_UP = f"\033[{len(applications)}A"
    # LINE_CLEAR = "\x1b[2K"

    # while True:
    #     print_application_status(applications, True, next(spinner))
    #     time.sleep(0.25)
    #     print(LINE_UP, end=LINE_CLEAR)


def open_argocd_port_forward():
    """
    Create an port forward to the ArgoCD server service, and open that page in a web browser
    """
    # TODO remove the password stuff, since we can assume we're running on the network
    result = subprocess.run(
        "kubectl -n argocd get secret argocd-initial-admin-secret -o jsonpath={{.data.password}} | base64 -d".format(),
        shell=True,
        capture_output=True,
    )
    subprocess.run(
        f"echo {result.stdout.decode('utf-8')} | pbcopy",
        shell=True,
        capture_output=True,
    )
    # Open a new tab in the default browser to localhost:8080 with webbrowser
    webbrowser.open_new_tab("http://localhost:8080/argocd")

    # Port forward to the ArgoCD UI
    subprocess.run(
        "kubectl -n argocd port-forward svc/argocd-server -n argocd 8080:443",
        shell=True,
    )


def print_application_status(applications, show_updates=False, spin_char="-") -> None:
    """
    Iterate over a list of applications, getting their status, and then printing the name. Green means the application
    is up fully, yellow means the application is not healthy, but hasn't failed, and red means the application has
    failed. Also can print a spinner if show_updates is True.

    Args:
        applications (list): A list of application names
        show_updates (bool): Whether to print a spinner for each application
        spin_char (str): The character to use for the spinner
    """
    v1 = client.CustomObjectsApi()

    for app in applications:
        app = v1.get_namespaced_custom_object(
            group="argoproj.io",
            version="v1alpha1",
            name=app,
            namespace="argocd",
            plural="applications",
        )

        if "operationState" in app["status"]:
            phase = app["status"]["operationState"]["phase"]
        else:
            phase = "Failed"
        status = app["status"]["health"]["status"]

        if phase == "Failed":
            print(colored(app["metadata"]["name"], "red"), end=" ")
        elif status != "Healthy":
            print(colored(app["metadata"]["name"], "yellow"), end=" ")
        else:
            print(colored(app["metadata"]["name"], "green"), end=" ")

        if show_updates:
            if phase == "Failed" or status != "Healthy":
                print(spin_char)
            else:
                print("✔")
        else:
            print()


def get_tracked_applications():
    """
    Get all applications under the master stack application. It loops over the resources field, picking out names
    whose kind is Application
    """
    api = client.CustomObjectsApi()

    master = api.get_namespaced_custom_object(
        group="argoproj.io",
        version="v1alpha1",
        name="master-stack",
        namespace="argocd",
        plural="applications",
    )

    resources = master["status"]["resources"]

    application_names = []
    for resource in resources:
        if resource["kind"] == "Application":
            application_names.append(resource["name"])

    return application_names


def _get_argocd_bearer_token() -> str:
    """
    Get a bearer token for authenticating with the ArgoCD API. Takes no arguments, because we get a username and
    password for ArgoCD from environment variables
    Return:
        token (str): A bearer token for authenticating with the ArgoCD API.
    """
    # Get an auth token, using the configured
    auth_response = requests.post(
        f"{settings.argocd_url}/api/v1/session",
        data=json.dumps(
            {"username": settings.argocd_username, "password": settings.argocd_password}
        ),
        verify=False,
    )

    token = auth_response.json()["token"]
    return token


def _get_master_stack_application() -> dict:
    """
    Get an object representing the master-stack application, located in the ArgoCD namespace
    Return:
        api_response (object): An object representing the master-stack application, located in the ArgoCD namespace
    """
    try:
        api_instance = client.CustomObjectsApi()

        # Get the master-stack application, located in the ArgoCD namespace. This is what should be creating all of
        # our applications, so we want disable the target application in master-stacks parameters, so it doesn't
        # recreate it
        namespace = "argocd"
        name = "master-stack"
        group = "argoproj.io"
        version = "v1alpha1"
        plural = "applications"
        api_response = api_instance.get_namespaced_custom_object(
            group, version, namespace, plural, name
        )

        return api_response
    except Exception as e:
        logging.error("Failed to get master-stack application, see error below")
        print(e)
    return {}


def _is_application_active(application: str, resources: dict) -> bool:
    """
    Check whether an application is active, i.e. whether or not it is in the list of resources created by master stack
    Args:
        application (str): The name of the application to check
        resources (list): A list of resources created by the master stack
    Return:
        bool: True if an application with the name matching the 'application' argument exists, and false otherwise
    """
    for resource in resources:
        if (
            resource.get("kind") == "Application"
            and resource.get("name") == application
        ):
            return True
    return False


def _is_application_enabled_in_master_stack(
    application: str, application_dict: dict
) -> bool:
    """
    Check if an application is enabled in master stack, i.e. the value 'application'.enabled is set.
    By default, its true, so it should only exist if its set to false.
    Args:
        application (str): The name of the application to check
        application_oject (object): The object representing the application, as returned by the ArgoCD API
    Return:
        bool: True if the application is enabled in master stack, and false otherwise.
    """
    source_helm_params = _get_source_helm_params(application_dict)
    for variable in source_helm_params:
        if application in variable.get("name"):  # type: ignore
            return False
    return True


def _get_source_helm_params(api_response: dict) -> list[dict[str, str]]:
    """
    Get the helm parameters from the api_response object. Will throw an error if no helm parameters are found.
    Args:
        api_response (object): The object representing the application, like a YAML output
    Return:
        source_helm_params (object): The helm parameters from the api_response object.
    """
    source_helm_params = (
        api_response.get("spec", {})
        .get("source", {})
        .get("helm", {})
        .get("parameters", [])
    )

    if not source_helm_params:
        logging.error("No helm parameters found")
        sys.exit(1)
    else:
        return source_helm_params


def _install_master_stack_if_not_present(environment: str) -> None:
    """
    Checks if an application named master-stack exists in the 'argocd' application, and if it doesn't,
    fill out the master-stack application template, and install it to the ArgoCD namespace
    Args:
        environment (str): The name of the environment we want to install into. For now, the valid options are
        dev-laptop, dev-cluster, and prod-cluster. TODO Update this documentation to reflect our new naming
        convention:
        qa-cluster: The K3s cluster running on our laptop
        preprod-cluster: The RKE cluster made out of VMs we use for testing in a working environment
        prod-cluster: The RKE cluster made out of VMs used to run our actual services
    """
    custom_objects_api = client.CustomObjectsApi()
    # Check whether an argocd application named 'master-stack' exists...
    try:
        custom_objects_api.get_namespaced_custom_object(
            group="argoproj.io",
            version="v1alpha1",
            namespace="argocd",
            plural="applications",
            name="master-stack",
        )
    # ...and if not, create it.
    except client.exceptions.ApiException as e:
        if e.status == 404:
            master_app_string = template_utils.template_file(
                "master-app.yaml", {"environment": environment}
            )
            master_app_dict = yaml.safe_load(master_app_string)
            logging.info("No application named 'master-stack' exists, creating it")
            custom_objects_api.create_namespaced_custom_object(
                group="argoproj.io",
                version="v1alpha1",
                namespace="argocd",
                plural="applications",
                body=master_app_dict,
                pretty=True,
            )
        else:
            logging.info(
                "Master-stack already exists. If you want to force a recreation, use"
                + " --force-master-reconfiguration[Not yet working]"
            )


def _wait_for_argocd_pods_to_be_ready() -> None:
    """
    Wait for all the ArgoCD pods to be up and healthy. We'll also print them with a
    fun spinner
    """
    # TODO Make timeout configurable
    v1 = client.CoreV1Api()
    timeout = time.time() + 60 * 5  # 5 minutes from now
    spinner = itertools.cycle(["-", "\\", "|", "/"])

    done_waiting = False

    # Check if all the pods are ready in a loop. If they aren't, we just print a waiting message with a spinner
    # This will time out after five minutes
    while not done_waiting:
        pods = v1.list_namespaced_pod("argocd")
        all_healthy = True
        for pod in pods.items:
            if pod.status.phase != "Running":
                all_healthy = False
                break
        if all_healthy:
            logging.info("Argocd pods are ready, installing master-stack")
            done_waiting = True
        if time.time() > timeout:
            logging.error(
                "Timeout reached, exiting. This is not an error with atils, there is likely something going"
                + " on in the cluster"
            )
            exit(1)

        print("Waiting for argocd pods to come up healthy " + next(spinner), end="\r")
        time.sleep(0.25)



import hvac
import logging

from hvac.api.auth_methods import Kubernetes

# TODO make it so that logging is set up using config stored in config.py
logging.basicConfig(level=logging.DEBUG)


def setup_client() -> hvac.Client:
    client = hvac.Client("http://vault.dev.local/")
    if client.is_authenticated:
        return client
    else:
        logging.error("Failed to authenticate with Vault")
        exit(1)


def setup_kubernetes_auth():
    client = setup_client()

    auth_methods = client.sys.list_auth_methods()

    if "kubernetes/" not in auth_methods["data"].keys():
        client.sys.enable_auth_method(
            method_type="kubernetes",
            path="kubernetes",
        )

    Kubernetes(client.adapter).create_role("argocd", ["argocd-repo-server"], ["argocd"])


setup_kubernetes_auth()



import logging

import yaml
from deepdiff import DeepDiff

logging.basicConfig(level=logging.DEBUG)


def compare_yaml_strings(yaml_str_1: str, yaml_str_2: str):
    """
    Given two string representing valid yaml, convert them to dictionaries and compare using
    deepdiff. Return false if there are differences, true if there are no differences.
    Args:
        yaml_str_1 (str): A string representing a yaml object
        yaml_str_2 (str): A string representing a yaml object
    Returns:
        bool: True if there are no differences, False if there are differences.
    """
    yaml_dict_1 = yaml.safe_load(yaml_str_1)
    yaml_dict_2 = yaml.safe_load(yaml_str_2)

    return _compare_yaml_dicts(yaml_dict_1, yaml_dict_2)


def compare_yaml_files(yaml_file_1: str, yaml_file_2: str) -> bool:
    """
    Given two strings that represent the paths to yaml files, load their contents as strings,
    convert them to dictionaries and compare using deepdiff. Return false if there are differences,
    true if there are no differences.
    Args:
        yaml_str_1 (str): A string representing a yaml object
        yaml_str_2 (str): A string representing a yaml object
    Returns:
        bool: True if there are no differences, False if there are differences.
    """
    with open(yaml_file_1, "r") as f1, open(yaml_file_2, "r") as f2:
        yaml_file_1_contents = f1.read()
        yaml_file_2_contents = f2.read()

    yaml_dict_1: dict = yaml.safe_load(yaml_file_1_contents)
    yaml_dict_2: dict = yaml.safe_load(yaml_file_2_contents)

    return _compare_yaml_dicts(yaml_dict_1, yaml_dict_2)


def find_item_index(search_level: str, item_name: str, yaml_data: dict) -> int:
    """
    Given a python object representing a kubeconfig file, a search level representing a resource type
    to search in, and an item name, return the index of the item in the yaml_data object.
    Args:
        search_level (str): The resource type to search in
        item_name (str): The name of the item to search for
        yaml_data (object): The python object representing the kubeconfig file
    Returns:
        int: The index of the item in the yaml_data object, or -1 if the item is not found.
    """
    for i, item in enumerate(yaml_data[search_level]):
        if item["name"] == item_name:
            return i
    return -1


def merge_kubeconfigs(
    kubeconfig1: str,
    kubeconfig2: str,
    kubeconfig1_cluster_name: str,
    kubeconfig2_cluster_name: str,
    kubeconfig1_user_name: str,
    kubeconfig2_user_name: str,
    kubeconfig1_context_name: str,
    kubeconfig2_context_name: str,
) -> str:
    """
    Given two kubeconfigs, and the names of the clusters and users to copy, merge the second kubeconfig into the first.
    Args:
        kubeconfig1 (str): The path to the first kubeconfig file
        kubeconfig2 (str): The path to the second kubeconfig file
        kubeconfig1_cluster_name (str): The name of the cluster in the first kubeconfig to copy to
        kubeconfig2_cluster_name (str): The name of the cluster in the second kubeconfig to copy from
        kubeconfig1_user_name (str): The name of the user in the first kubeconfig to copy to
        kubeconfig2_user_name (str): The name of the user in the second kubeconfig to copy from
        kubeconfig1_context_name (str): The name of the context in the first kubeconfig to copy to
        kubeconfig2_context_name (str): The name of the context in the second kubeconfig to copy from
    Returns:
        str: The path to the merged kubeconfig file.
    """
    # Load the YAML data from both kubeconfig files
    with open(kubeconfig1, "r") as f1, open(kubeconfig2, "r") as f2:
        kubeconfig_data1: dict = yaml.safe_load(f1)
        kubeconfig_data2: dict = yaml.safe_load(f2)

    # Cluster copying section
    data_1_index = find_item_index(
        "clusters", kubeconfig1_cluster_name, kubeconfig_data1
    )
    data_2_index = find_item_index(
        "clusters", kubeconfig2_cluster_name, kubeconfig_data2
    )

    kubeconfig_data1 = _replace_item_in_yaml_array(
        kubeconfig_data1,
        "clusters",
        data_1_index,
        kubeconfig_data2,
        "clusters",
        data_2_index,
    )

    # Copy the user section from kubeconfig2 to kubeconfig1
    data_1_index = find_item_index("users", kubeconfig1_user_name, kubeconfig_data1)
    data_2_index = find_item_index("users", kubeconfig2_user_name, kubeconfig_data2)

    kubeconfig_data1 = _replace_item_in_yaml_array(
        kubeconfig_data1, "users", data_1_index, kubeconfig_data2, "users", data_2_index
    )

    # Copy the context section from kubeconfig2 to kubeconfig1
    data_1_index = find_item_index(
        "contexts", kubeconfig1_context_name, kubeconfig_data1
    )

    data_2_index = find_item_index(
        "contexts", kubeconfig2_context_name, kubeconfig_data2
    )

    kubeconfig_data1 = _replace_item_in_yaml_array(
        kubeconfig_data1,
        "contexts",
        data_1_index,
        kubeconfig_data2,
        "contexts",
        data_2_index,
    )

    # Serialize the combined kubeconfig data and return it as a string
    return yaml.dump(kubeconfig_data1)


def _compare_yaml_dicts(yaml_dict_1: dict, yaml_dict_2: dict) -> bool:
    """
    Given two dictionaries representing yaml files, compare them using deepdiff. Return false if there are differences,
    true if there are no differences.
    Args:
        yaml_dict_1 (dict): A dictionary representing a yaml object
        yaml_dict_2 (dict): A dictionary representing a yaml object
    Returns:
        bool: True if there are no differences, False if there are differences.
    """
    diff = DeepDiff(yaml_dict_1, yaml_dict_2, ignore_order=True)

    if len(diff) > 0:
        return False
    else:
        return True


def _replace_item_in_yaml_array(
    yaml_dict_1: dict,
    yaml1_array_name: str,
    yaml1_array_index: int,
    yaml_dict_2: dict,
    yaml2_array_name: str,
    yaml2_array_index: int,
) -> dict:
    """
    Given two dictionaries representing yaml files, replace or append the item at yaml_dict_1[yaml1_array_name]
    [yaml1_array_index] with the item at yaml_dict_2[yaml2_array_name][yaml2_array_index]. Return yaml_dict_1
    with the items replaced, or appended

    Args:
        yaml_dict_1: The dictionary representation of the target yaml file
        yaml1_array_name: The name of the array at the top level of the yaml file to replace
        yaml1_array_index: The index of the item in the array to replace
        yaml_dict_2: The dictionary representation of the source yaml file
        yaml2_array_name: The name of the array at the top level of the yaml file to use as a source
        yaml2_array_index: The index of the item in the array to use as a source

    Returns:
        dict: The dictionary representation of the target yaml file with the items replaced, or appended.
    """
    if yaml1_array_index > -1:
        yaml_dict_1[yaml1_array_name][yaml1_array_index] = yaml_dict_2[
            yaml2_array_name
        ][yaml2_array_index]
    else:
        yaml_dict_1[yaml1_array_name].append(
            yaml_dict_2[yaml2_array_name][yaml2_array_index]
        )
    return yaml_dict_1



import logging
import os

from atils.common.settings import settings


def get_full_atils_dir(dir_name: str) -> str:
    """
    Given the name of a subdirectory, return the result of using os.join
    on that directory and the directory where our monorepo is installed
    Args:
        dir_name (str): The name of the subdirectory to join
    Returns:
        str: The full path to the subdirectory
    """
    return os.path.join(settings.INSTALL_DIR, settings[dir_name])


def get_logging_level() -> int:
    """
    Return the logging level configured. We use this function so we return the
    constants in the logging module, rather than a string we need to parse anyway
    """
    if settings.LOG_LEVEL == "DEBUG":
        return logging.DEBUG
    elif settings.LOG_LEVEL == "INFO":
        return logging.INFO
    elif settings.LOG_LEVEL == "WARNING":
        return logging.WARNING
    elif settings.LOG_LEVEL == "ERROR":
        return logging.ERROR
    elif settings.LOG_LEVEL == "CRITICAL":
        return logging.CRITICAL
    else:
        logging.warning("No log level set, defaulting to INFO")
        return logging.INFO



import argparse
import base64
import logging
import os
import shutil
import sys

from atils.common import config, yaml_utils
from atils.common.settings import settings
from kubernetes import client
from kubernetes import config as k8s_config

client.rest.logger.setLevel(logging.WARNING)

logging.basicConfig(level=config.get_logging_level())  # type: ignore


def main(args: list[str]):
    # This variable tracks whether or not we have configuration available to run kubernetes commands
    CAN_RUN: bool = load_config()

    if not CAN_RUN:
        logging.error("No configuration available to run kubernetes commands")
        exit(1)

    # TODO add autocomplete for secret names
    parser: argparse.ArgumentParser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        help="Select a subcommand", dest="subparser_name"
    )

    # Options for managing kubernetes secrets. This is different from Vault secrets management
    secrets_parser = subparsers.add_parser(
        "secrets", help="Commands to manage kubernetes secrets"
    )

    secrets_parser.add_argument(
        "command", choices=["decode"], help="Which command to use to operate on secrets"
    )

    secrets_parser.add_argument(
        "secret_name", help="The name of the secret to operate on"
    )

    secrets_parser.add_argument(
        "-n", "--namespace", help="The namespace of the secret to operate on"
    )

    # Options for RKE cluster setup
    rke_parser = subparsers.add_parser(
        "rke-setup", help="Commands to manage the rke installation"
    )
    rke_parser.add_argument(
        "-f",
        "--force",
        help="force the recreation of a cluster, in cases where it's already running",
    )
    rke_parser.add_argument(
        "-rk",
        "--replace-kubeconfig",
        help="Copies a kubeconfig generated by rke to the system-wide kubeconfig",
        action="store_true",
        default=False,
    )
    rke_parser.add_argument(
        "cluster_name",
        help="the name of the cluster to set up. This should match an rke file",
    )

    if len(args) == 0:
        parser.print_help(sys.stderr)
        sys.exit(1)

    arguments: argparse.Namespace = parser.parse_args(args)

    if arguments.subparser_name == "rke-setup":
        if vars(arguments).get("replace_kubeconfig"):
            merge_and_replace_kubeconfig(arguments.cluster_name)
        else:
            setup_rke_cluster(
                arguments.cluster_name,
            )

    elif arguments.subparser_name == "secrets":
        args_dict: dict[str, str] = vars(arguments)
        if args_dict["command"] == "decode":
            if args_dict["secret_name"] is None:
                logging.error("Error: A secret name must be provided ")
                sys.exit(1)
            else:
                get_and_decode_secret(args_dict["secret_name"], args_dict["namespace"])


def load_config() -> bool:
    """
    Loads some kind of kubernetes configuration, either in-cluster or in kubeconfig,
    and returns true if it is able to. If it cannot, it will return false, and the main
    function is in charge of failing if it needs to

    Returns:
        (bool): True if a config could be loaded, false otherwise
    """
    # TODO also support running from a pod
    if "KUBECONFIG" in os.environ:
        k8s_config.load_kube_config(os.environ["KUBECONFIG"])
        return True
    # Check if ~/.kube/config exists
    if os.path.exists(os.path.expanduser("~/.kube/config")):
        k8s_config.load_kube_config()
        return True
    else:
        logging.info(
            "Could not find a kubeconfig. Any commands related to kubernetes have been disabled"
        )
        return False


def merge_and_replace_kubeconfig(cluster_name: str) -> None:
    """
    Merge a kubeconfig generated by rke with the system-wide kubeconfig. This is done by creating a backup of the
    system-wide kubeconfig, then merging the two kubeconfigs, and then replacing the system-wide kubeconfig with the
    merged kubeconfig. This is done to avoid conflicts with other kubeconfigs that may be present in the system-wide
    kubeconfig. This function is used to replace the system-wide kubeconfig with a kubeconfig generated by rke.
    Args:
        cluster_name (str): The name of the cluster. For now, the valid options are
        dev-laptop, dev-cluster, and prod-cluster. TODO Update this documentation to reflect our new naming
        convention:
        qa-cluster: The K3s cluster running on our laptop
        preprod-cluster: The RKE cluster made out of VMs we use for testing in a working environment
        prod-cluster: The RKE cluster made out of VMs used to run our actual services
    """
    shutil.copy(
        f"{settings.KUBECONFIG_LOCATION}/config",
        f"{settings.KUBECONFIG_LOCATION}/config.bak",
    )
    merged_kubeconfig = yaml_utils.merge_kubeconfigs(
        f"{settings.KUBECONFIG_LOCATION}/config",
        os.path.join(
            settings.SCRIPT_INSTALL_DIRECTORY,
            f"../kubernetes/rke/kube_config_{cluster_name}.yaml",
        ),
        # TODO This is still hard coded
        "kind-dev-laptop",
        "local",
        "kind-dev-laptop",
        "kube-admin-local",
        "kind-dev-laptop",
        "local",
    )

    with open(f"{settings.KUBECONFIG_LOCATION}/config", "w") as kfile:
        kfile.truncate(0)
        kfile.write(merged_kubeconfig)


def setup_rke_cluster(cluster_name: str, force: bool = False):
    """
    Set up an RKE whose specification is in kubernetes/rke/cluster_name.yaml. First attempts to check if the cluster is
    available TODO this doesn't work properly yet.
    Args:
        cluster_name (str): The name of the cluster. For now, the valid options are
        dev-laptop, dev-cluster, and prod-cluster. TODO Update this documentation to reflect our new naming
        convention:
        qa-cluster: The K3s cluster running on our laptop
        preprod-cluster: The RKE cluster made out of VMs we use for testing in a working environment
        prod-cluster: The RKE cluster made out of VMs used to run our actual services
        force (bool): If True, the cluster will be set up even if it is already running. If False, the cluster will
        not be set up if it is already running. Defaults to False.
    """
    cluster_availability = _check_cluster_availability()

    if cluster_availability and not (force):
        logging.error(
            "Cluster was reachable and --force was not used. If you wish to overwrite your cluster, use --force."
        )
        exit(0)
    elif not (cluster_availability) or force:
        # TODO Make this take config, and try to move these configuration files
        cluster_config_location: str = os.path.join(
            settings.SCRIPT_INSTALL_DIRECTORY, f"../kubernetes/rke/{cluster_name}.yaml"
        )
        if os.path.isfile(cluster_config_location):
            # TODO we should add our own error checking here, rather than relying on RKE's.
            os.system(f"rke up --config {cluster_config_location}")
            merge_and_replace_kubeconfig(cluster_name)
        else:
            logging.error(
                f"Could not find an rke file named {cluster_name}.yaml. Aborting"
            )
            exit(1)


def check_namespace_exists(namespace_name: str) -> bool:
    """
    Check if a given kubernetes namespace exists
    Args:
        namespace_name (str): The name of the namespace to check
    Returns:
        bool: True if the namespace exists, False otherwise.
    """
    # Create Kubernetes API client
    v1 = client.CoreV1Api()

    # Use the API client to list all namespaces
    namespace_list = v1.list_namespace().items

    # Check if the namespace exists
    if any(namespace.metadata.name == namespace_name for namespace in namespace_list):
        return True
    else:
        return False


def get_and_decode_secret(secret_name: str, secret_namespace: str) -> None:
    """
    Get a kubernetes secret, and then decode and pretty print it
    Args:
        secret_name (str): The name of the secret to decode and print
        secret_namespace (str): The namespace the given secret is located in
    """
    try:
        # Create a Kubernetes API client
        api = client.CoreV1Api()

        if secret_namespace is None:
            current_context = k8s_config.list_kube_config_contexts()[1]
            # Check if current_context["context"]["namespace"] is None
            secret_namespace = current_context.get("context").get(
                "namespace", "default"
            )

        # Get the Secret from the specified namespace
        secret = api.read_namespaced_secret(
            name=secret_name, namespace=secret_namespace
        )

        terminal_width = shutil.get_terminal_size().columns

        print("=" * int(terminal_width / 2))
        print(secret.metadata.name.center(int(terminal_width / 2)))
        print("=" * int(terminal_width / 2))

        # Decode and pretty print the data items
        for key, value in secret.data.items():
            decoded_value = base64.b64decode(value).decode("utf-8")
            # If decoded_value is more than one line, print on a new line
            if "\n" in decoded_value:
                print(f"{key}:")
                print(decoded_value)
            else:
                print(f"{key}: {decoded_value}")

            print("=" * int(terminal_width / 2))

    except Exception as e:
        logging.error(f"An error occurred: {e}")


def get_current_namespace() -> str:
    """
    Gets the current namespace set in the current context
    Returns:
        str: The current namespace set in the current context
    """
    contexts, active_context = k8s_config.list_kube_config_contexts()

    if "namespace" in active_context["context"].keys():
        return active_context["context"]["namespace"]
    else:
        return "default"


# TODO after we've standardized the names for our cluster, we should make this
def _check_cluster_availability() -> bool:
    """
    Checks if the cluster in the active context is available, by attempting to list nodes on a five second timeout
    Returns:
        true if the operation to list nodes succeeds, and false otherwise
    """
    try:
        contexts, active_context = k8s_config.list_kube_config_contexts()
        cluster_name: str = active_context["context"]["cluster"]
        logging.debug(f"Checking cluster availability for cluster: {cluster_name}")
        # Create a CoreV1Api instance
        api_instance = client.CoreV1Api()

        # Get the list of nodes in the cluster
        api_response = api_instance.list_node(timeout_seconds=5)

        # Check if the cluster is reachable
        if api_response:
            logging.info(f"Cluster '{cluster_name}' is reachable.")
            return True
        else:
            logging.info(f"Cluster '{cluster_name}' is not reachable.")
            return False
    except Exception as e:
        if "[Errno 64]" in e.args[0]:
            logging.error(
                "Cluster host is down. If you are trying to setup a new cluster, use the"
                + ' "rke-setup [cluster_name]" subcommand'
            )
            return False
        elif "Max retries execeeded with url:" in e.args[0]:
            logging.error(
                "Cluster host is unreachable. If you are trying to setup a new cluster, use the"
                + ' "rke-setup [cluster_name]" subcommand'
            )
            return False

        else:
            logging.error("We found a new kind of issue! This is a bug.")
            logging.error(e)
            return False



import os

from jinja2 import Environment, FileSystemLoader, FunctionLoader
from pkg_resources import resource_string


def load_from_self(filename: str) -> str:
    """
    Given the name of a template file in the atila/templates directory, load the
    contents of the file as a bytes object

    Args:
        filename (str): The name of the file to load

    Returns:
        bytes: The contents of the file as a bytes object
    """
    return resource_string("atils.templates", filename).decode("utf-8")


def template_file(template_path: str, args_dict: dict) -> str:
    """
    Given the path to a template file, load the template and substitute the
    values in the dictionary into the template.

    Args:
        template_path (str): The path to the template file
        args_dict (dict): A dictionary of values to substitute into the template

    Returns:
        str: The substituted template content as a string.
    """
    env = Environment(loader=FunctionLoader(load_from_self))
    template = env.get_template(os.path.basename(template_path))
    substituted_content = template.render(args_dict)

    return substituted_content


def template_external_file(template_path: str, args_dict: dict) -> str:
    """
    Given the path to a template file that is not bundled with atils,load
    the template and substitute the values in the dictionary into the template.

    Args:
        template_path (str): The path to the template file
        args_dict (dict): A dictionary of values to substitute into the template

    Returns:
        str: The substituted template content as a string.
    """
    env = Environment(loader=FileSystemLoader(os.path.dirname(template_path)))
    template = env.get_template(os.path.basename(template_path))
    substituted_content = template.render(args_dict)

    return substituted_content


def template_file_and_output(
    template_path: str, output_path: str, args_dict: dict
) -> None:
    """
    Given a path to a template file, and arguments, template the file and then output
    it to a file

    Args:
        template_path (str): The path to the template file
        output_path (str): The path to the output file
        args_dict (dict, optional): A dictionary of values to substitute into the template. Defaults to None.
    """
    substituted_content = template_file(template_path, args_dict)

    with open(output_path, "w") as output_file:
        output_file.write(substituted_content)



from dynaconf import Dynaconf

# We're just setting up a dynaconf settings object, nothing super complex
settings = Dynaconf(
    envvar_prefix="ATILS",
    settings_files=["settings.yaml", ".secrets.yaml"],
    core_loaders=["YAML"],
)



import argparse
import json
import logging
import os
import subprocess

from atils.common import config

logging.basicConfig(level=config.get_logging_level())  # type: ignore


def main(args: list[str]):
    parser: argparse.ArgumentParser = argparse.ArgumentParser()

    parser.add_argument(
        "action",
        help="Select which action to perform. Defaults to build",
        default="build",
        nargs="?",
    )

    parser.add_argument(
        "--build-directory",
        "-bd",
        type=str,
        help="The directory where .atils_buildconfig.json is located",
    )

    parser.add_argument(
        "--actions-only", "-ao", help="Only show actions", action="store_true"
    )
    parser.add_argument(
        "--action-sets-only", "-aso", help="Only show action sets", action="store_true"
    )

    parser.add_argument(
        "--actions",
        nargs="*",
        help="Actions defined in .atils_buildconfig.json of current directory to run",
    )

    parser.add_argument(
        "--action-set",
        help="A single action set defined in .atils_buildconfig.json of current directory to run",
    )

    arguments: argparse.Namespace = parser.parse_args(args)

    directory: str = os.getcwd()
    show_actions: bool = True
    show_action_sets: bool = True

    if arguments.actions_only:
        show_actions = False
    if arguments.action_sets_only:
        show_action_sets = False

    if not show_actions and not show_action_sets:
        raise ValueError("You need to allow showing either actions or action sets")

    if arguments.build_directory is not None:
        directory = arguments.build_directory

    if arguments.action == "list":
        run_list_action(directory, show_actions, show_action_sets)
    else:
        if arguments.actions is not None:
            run_build_actions(arguments.actions, directory)
        elif arguments.action_set is not None:
            run_build_action_set(arguments.action_set, directory)
        else:
            default_action_set: str = _get_default_action_set(directory)
            if default_action_set:
                run_build_action_set(default_action_set, directory)
            else:
                logging.error(
                    "No default action set defined. Either define one in"
                    + f" {os.path.join(directory, '.atils_buildconfig.json')} or use an argument to specify what to do"
                )


def validate_listed_actions(
    available_actions: list[dict], listed_actions: list[str]
) -> None:
    """
    Takes a list of valid actions, a list of actions provided by a user, and ensures that all
    actions provided by the user are valid.
    Args:
        available_actions (list[object]): A list of objects representing available actions.
        listed_actions (list[str]): A list of strings representing actions provided by the user.
    Raises:
        ValueError: If any action provided by the user is not valid.
    Returns:
        None.
    """
    available_action_names = [action["name"] for action in available_actions]
    for action in listed_actions:
        if action not in available_action_names and action != "all":
            raise ValueError(f"{action} is not a valid action")


def run_list_action(directory: str, show_actions: bool, show_action_sets: bool) -> None:
    """
    List all available build actions, or only actions or only action sets.
    Args:
        directory (str): The directory where our .atils_buildconfig.json file to list is located
        show_actions (bool): Whether to show actions or not
        show_action_sets (bool): Whether to show action sets or not
    Raises:
        ValueError: If both show_actions and show_action_sets are True.
    Returns:
        None
    """
    available_actions = _get_available_actions(directory)
    available_action_sets = _get_available_action_sets(directory)

    if not show_actions and not show_action_sets:
        raise ValueError("Cannot specify both actions_only and action_sets_only")
    if show_actions:
        print("Actions")
        print("=======================================")
        for action in available_actions:
            _print_action(action)
    if show_action_sets:
        print("Action Sets")
        print("=======================================")
        for action_set in available_action_sets:
            _print_action_set(action_set, available_actions)


def run_build_action_set(action_set_name: str, directory: str) -> None:
    """
    Runs all the actions defined in a given action_set
    Args:
        action_set (dict): A dictionary representing an action set
        directory (str): The directory where our .atils_buildconfig.json file to list is located
    """
    available_actions: list[dict] = _get_available_actions(directory)
    action_set: dict = _get_action_set(action_set_name, directory)

    validate_listed_actions(available_actions, action_set["actions"])

    if "strict" in action_set and not action_set["strict"]:
        for action in available_actions:
            if action["name"] in action_set["actions"]:
                _run_action(action, directory)
    else:
        for action in available_actions:
            if action["name"] in action_set["actions"]:
                _run_action_strict(action, directory)


def run_build_actions(listed_actions: list[str], directory: str) -> None:
    # TODO cd to a directory, if one is specified
    # TODO allow us to mark if all build actions must pass
    """
    Run user-specified build actions, in the order specified by .atils_buildconfig.json.
    Args:
        listed_actions (list[str]): The list of actions provided by the user.
        directory (str): The directory to use as the current working directory.
    """
    available_actions = _get_available_actions(directory)

    validate_listed_actions(available_actions, listed_actions)

    if "all" in listed_actions:
        for action in available_actions:
            _run_action(action, directory)
    else:
        # Since I know you might ask... we do it this way to ensure that the actions are run in order.
        # The user can enter them arbitrarily, but available_actions is sorted, so we can use that as
        # our source of truth
        for action in available_actions:
            if action["name"] in listed_actions:
                _run_action(action, directory)


def _get_available_actions(directory: str) -> list[dict]:
    """
    List all actions available in a .atils_buildconfig.json file in the given directory.
    The file must be in the given directory and must be a valid JSON file.
    The file must contain a list of actions, each action must be an object with the following keys:
        - name: The name of the action
        - command: The command to run
        - order: The order in which to run the action. Lower numbers are run first.
    The actions are returned sorted by their order.
    Arguments:
        directory (str): The directory where the .atils_buildconfig.json file is located.
    Returns:
        A list of objects representing actions available in the .atils_buildconfig.json file.
    """
    filename: str = os.path.join(directory, ".atils_buildconfig.json")

    if os.path.isfile(filename):
        with open(filename) as f:
            data = json.load(f)
            if "actions" not in data:
                raise ValueError(f"No actions found in {filename}")
            return sorted(data["actions"], key=lambda x: x["order"])
    else:
        raise FileNotFoundError(f"{filename} does not exist")


def _get_available_action_sets(directory: str) -> list[dict]:
    """
    List all action sets available in a .atils_buildconfig.json file in the given directory.
    The file must be in the given directory and must be a valid JSON file.
    The file must contain a list of action sets, each action set must be an object with the following keys:
        - name: The name of the action set
        - actions: A list of actions in the action set
    The action sets are returned sorted by their name.
    Arguments:
        directory (str): The directory where the .atils_buildconfig.json file is located.
    Returns:
        A list of objects representing action sets available in the .atils_buildconfig.json file. Returns an empty
        list if no such list is found
    """
    filename = os.path.join(directory, ".atils_buildconfig.json")

    if os.path.isfile(filename):
        with open(filename) as f:
            data = json.load(f)
            if "action_sets" not in data:
                return []
            return sorted(data["action_sets"], key=lambda x: x["name"])
    else:
        raise FileNotFoundError(f"{filename} does not exist")


def _get_action_set(action_set: str, directory: str) -> dict:
    """
    Get a given action set from an .atils_buildconfig.json file
    Args:
        action_set (str): The name of the action set to get
        directory (str): The directory where the .atils_buildconfig.json file is located
    Returns:
        A dictionary representing the action set, or an empty dictionary if the action set is not found
    """
    available_action_sets = _get_available_action_sets(directory)

    for action_set_obj in available_action_sets:
        if action_set_obj["name"] == action_set:
            return action_set_obj
    raise ValueError(
        f"Action set {action_set} not available in {os.path.join(directory, '.atils_buildconfig.json')}"
    )


def _get_default_action_set(directory: str) -> str:
    """
    Given a directory with an .atils_buildconfig.json file, return the name of the default action set.
    Args:
        directory (str): The directory where the .atils_buildconfig.json file is located
    Returns:
        A string representing the name of the default action set, or an empty string if no default action set is found.
    """
    available_action_sets: list[dict] = _get_available_action_sets(directory)
    for action_set in available_action_sets:
        if "default" in action_set and action_set["default"]:
            return action_set["name"]
    return ""


def _print_action(action: dict) -> None:
    """
    Print the name and command of an action.
    Args:
        action (object): An object representing an action from a .atils_buildconfig.json file.
    """
    if "description" in action:
        print(f"{action['name']}: {action['command']} | {action['description']}")
    else:
        print(f"{action['name']}: {action['command']}")
    print()


def _print_action_set(action_set: dict, actions: list[dict]) -> None:
    """
    Print the name and actions of an action set.
    Args:
        action_set (dict): An object representing an action set from a .atils_buildconfig.json file.
        actions (list[dict]): A list of objects, representing all available actions
    """
    print(f"{action_set['name']}:")
    if "description" in action_set:
        print(f"  {action_set['description']}")
    print(action_set)
    for action in action_set["actions"]:
        if "description" in actions[action]:
            print(f"  {action} | {actions[action]['description']}")
        else:
            print(f"  {action} | {actions[action]['command']}")
    print()


def _run_action(action: dict, directory: str) -> None:
    """
    Run the command from an action, using subprocess.run
    Args:
        action (object): An object representing an action
    """

    _validate_action_can_run(action)
    subprocess.run(action["command"], shell=True, cwd=directory)


def _run_action_strict(action: dict, directory: str) -> None:
    """
    Run the command from an action, using subprocess.run. Fails if the command fails
    Args:
        action (object): An object representing an action
    """
    _validate_action_can_run(action)
    try:
        subprocess.run(action["command"], shell=True, cwd=directory, check=True)
    except subprocess.CalledProcessError as e:
        logging.error(f"Error running {action['name']}")
        exit(1)


def _validate_action_can_run(action: dict):
    """
    Given an action, checks if it is a CI-only action. If it is, and we are not running in a CI environment,
    exit the program.
    Args:
        action (dict): A dict representing an action from a .atils_buildconfig.json file.
    Returns:
        None.
    Raises:
        SystemExit: If we are not running in a CI environment and the action is a CI-only action.
    """
    if "ci_only" in action and action["ci_only"] and "ATILS_CI_ENV" not in os.environ:
        logging.error(
            f"Attempted to run a CI only-action ({action['name']}) in a non-CI environment. "
            + "If you know what you're doing, set ATILS_CI_ENV to true."
        )
        exit(1)



import sys

from atils import argocd
from atils import atils_kubernetes as kubernetes
from atils import build, helm, jobs


def main():
    if len(sys.argv) < 2:
        print("Atils requires at least one subcommand argument.")
        sys.exit(1)

    script_name: str = sys.argv[1]

    # TODO Make this use argparse. We'll get a proper help page up, and error
    # handling for incorrect arguments

    if script_name == "kubernetes":
        kubernetes.main(sys.argv[2:])
    elif script_name == "argocd":
        argocd.main(sys.argv[2:])
    elif script_name == "build":
        build.main(sys.argv[2:])
    elif script_name == "job":
        jobs.main(sys.argv[2:])
    elif script_name == "helm":
        helm.main(sys.argv[2:])
    else:
        print(f"Unrecognized subcommand: {script_name}")
        print("Valid subcommands are: kubernetes, argocd, build, job, helm")
        sys.exit(1)



import argparse
import itertools
import json
import logging
import os
import subprocess
import time

import yaml

from atils import atils_kubernetes
from atils.common import config, template_utils
from kubernetes import client
from kubernetes import config as k8s_config
from kubernetes import utils

client.rest.logger.setLevel(logging.ERROR)

logging.basicConfig(level=config.get_logging_level())  # type: ignore


def main(args: str) -> None:
    # This variable tracks whether or not we have configuration available to run kubernetes commands
    CAN_RUN: bool = atils_kubernetes.load_config()

    if not CAN_RUN:
        logging.error("No configuration available to run kubernetes commands")
        exit(1)

    parser: argparse.ArgumentParser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        help="Commands to manage kubernetes jobs", dest="subparser_name"
    )

    run_parser = subparsers.add_parser("run")
    run_parser.add_argument("job_name", help="Name of the job to run")
    # TODO Add some values for jobs, if needed, so we can set them with this argument
    run_parser.add_argument(
        "--set", help="Set values to fill in job template. WIP, not currently working"
    )
    run_parser.add_argument("--tag", help="Image tag to use for the job")

    pvc_parser = subparsers.add_parser("manage-pvc")
    pvc_parser.add_argument(
        "--pvc-name", "-pn", help="The name of the PVC to launch a management pod for"
    )
    pvc_parser.add_argument(
        "--namespace",
        "-n",
        help="The namespace the PVC is located in. Defaults to current namespace",
    )

    list_parser = subparsers.add_parser("list")

    arguments: argparse.Namespace = parser.parse_args(args)

    if arguments.subparser_name == "run":
        job_args = {}

        args_dict = vars(args)
        if "image" in args_dict and args_dict["tag"] is not None:
            job_args["image_tag"] = args_dict["tag"]
        else:
            job_args["image_tag"] = "latest"

        run_job_cli(arguments.job_name, job_args)
    elif arguments.subparser_name == "manage-pvc":
        args_dict = vars(arguments)
        current_namespace = atils_kubernetes.get_current_namespace()

        if "namespace" in args_dict.keys():
            if args_dict.get("namespace") is not None:
                launch_pvc_manager(args_dict["pvc_name"], args_dict["namespace"])
            else:
                launch_pvc_manager(args_dict["pvc_name"], current_namespace)
        else:
            launch_pvc_manager(args_dict["pvc_name"], current_namespace)
    elif arguments.subparser_name == "list":
        list_available_jobs()
    else:
        logging.error(f"Unrecognized command {arguments.subparser_name}")
        exit(1)


def launch_pvc_manager(pvc_name: str, namespace: str) -> None:
    """
    Launch the PVC manager job for a pod. This pod just waits for an hour, so you can connect
    to it
    Args:
        pvc_name (str): The name of the PVC to modify and manage
        namespace (str): The namespace the PVC is located in
    """
    if pvc_name is None:
        logging.error("--pvc-name must be provided")
        exit(1)
    else:
        _scale_down_controller_from_pvc(pvc_name)

        # TODO split out the functionality of rendering the job, waiting, and launching our exec
        with client.ApiClient() as api_client:
            # Launch the pod, probably with a kubectl command
            rendered_job = _render_job(
                "pvc-manager-interactive", {"pvc_name": pvc_name}
            )
            _launch_job(rendered_job)

            logging.info("Waiting 5 seconds for the job pod to come up")
            time.sleep(5)

            api_instance = client.CoreV1Api(api_client)
            pods = api_instance.list_namespaced_pod(
                namespace,
                # TODO This is hardcoded
                label_selector="job-name=pvc-manager-pvc-jellyfin",
                field_selector="status.phase!=Succeeded",
            )
            if len(pods.items) < 1:
                logging.error("Was not able to find any running pods for this job")
                exit(1)
            else:
                pod_name = pods.items[0].metadata.name
                exec_command = f"kubectl exec -it {pod_name} -- /bin/sh"

                subprocess.run(
                    f"echo {exec_command} | pbcopy",
                    shell=True,
                    capture_output=True,
                )
                print("Copied kubectl exec command to clipboard")

            # TODO I'd like to have this automatically open the pod, but that's a pain right now
            # I think we're going to want to use the CMD module
            # response = stream(
            #     core_api_client.connect_get_namespaced_pod_exec(
            #         "qbittorrent-68cf455bb9-nzpvb",
            #         namespace=namespace,
            #         # container=rendered_job["spec"]["template"]["spec"]["containers"][0]["name"],
            #         container="qbittorrent",
            #         stdin=True,
            #         stdout=True,
            #         stderr=True,
            #         command="/bin/sh",
            #     )
            # )

            # while response.is_open():
            #     response.update(timeout=1)
            # if response.peek_stdout():
            #     print(f"STDOUT: {response.read_stdout()}")
            # if response.peek_stderr():
            #     print(f"STDERR: {response.read_stderr()}")
            # if commands:
            #     c = commands.pop(0)
            #     print(f"Running command... {c}\n")
            #     response.write_stdin(c + "\n")
            # else:
            #     break

        # Scale the controller back up
        # if controller_name is not None:
        #     _modify_controller_replicas(
        #         controller_name,
        #         controller_namespace,
        #         controller_kind,
        #         previous_replicas,
        #     )


def list_available_jobs() -> None:
    # TODO exclude the docs directory, include a list of valid arguments
    """
    Print all the jobs available to run in the jobs directory to the console
    """
    jobs_dir = config.get_full_atils_dir("JOBS_DIR")

    root, dirs, files = next(os.walk(jobs_dir))
    for job in dirs:
        description = "No description provided"
        description_location = os.path.join(jobs_dir, job, "description.txt")
        if os.path.exists(description_location):
            with open(description_location) as file:
                description = file.read()
                if len(description) > 250:
                    description = description[0:251] + "..."

        print(f"{job}:      {description}")


def run_job_cli(job_name: str, args=None) -> None:
    """
    Given a job name and list of args, render the job template, then run the job
    Args:
        job_name (str): The name of the job to run. Must be a directory in the JOBS_DIR directory
        args (dict[str, str]): A dictionary representing arguments. Each key should correspond to a
        variable in a job template, with each value representing what should be filled in
    """
    rendered_job = _render_job(job_name, args)
    _launch_job(rendered_job)
    logging.info(f"Job {job_name} created")


def _clear_job_name(job_name: str, namespace: str) -> None:
    """
    We don't do a GenerateName for our jobs, so we need to make sure that the generated job name is available.
    So given a job name, and a namespace, delete the job, and then make sure it's deleted before letting us out
    """
    # Get all the jobs in the namespace, and then loop over them, looking for a matching name field
    # If found, we'll then delete the job, and wait for it to clear out
    v1 = client.BatchV1Api()
    for job in v1.list_namespaced_job(namespace).items:
        if job.metadata.name == job_name:
            # TODO Let's also delete all pods associated with the job
            # TODO the best way to do that is going to be to try and get a pod with all the matching labels,
            # so let's just refactor to not be afraid of error handling
            v1.delete_namespaced_job(name=job_name, namespace=namespace)
            # Wait until the job is deleted
            dots = itertools.cycle([".  ", ".. ", "..."])
            spinner = itertools.cycle(["-", "\\", "|", "/"])

            # TODO split out the waiting logic
            job = v1.read_namespaced_job(name=job_name, namespace=namespace)
            while job:
                try:
                    job = v1.read_namespaced_job(name=job_name, namespace=namespace)
                    print(
                        f"Waiting for job {job_name} to be deleted{next(dots)} {next(spinner)}",
                        end="\r",
                    )
                    time.sleep(0.2)
                except client.rest.ApiException as e:
                    if e.status == 404:
                        job = None
                    else:
                        raise e
            print("\n")
            logging.info(f"Job {job_name} deleted")
            return
    logging.info(f"No job named {job_name} found in namespace {namespace}")


def _get_controller_from_pvc(
    pvc_name: str, namespace: str = ""
) -> tuple[str, str, str]:
    """
    Given the name of a PVC, retrieve the pod it's attached to, and then from that, retrieve it's controller.
    We can then scale down that controller, so our job pod can mount it
    Args:
        pvc_name (str): The name of the PVC whose pod's controller we want to find
        namespace (str): The namespace the PVC is located. If no namespace is provided, then it uses the current
        namespace
    Returns:
        tuple[str, str, str]: If a pod is found for the PVC, returns the pod controller's name, its namespace, and the
        controller kind
    """
    # TODO Validate that the PVC exists, so we can use it on PVCs without a pod
    description: str = _get_pvc_description(pvc_name, namespace)

    pod_namespace: str = ""
    if namespace:
        pod_namespace = namespace

    pod_info = _get_information_from_description(description)

    if "used_by" not in pod_info.keys():
        logging.error(
            f"Could not find a pod for {pvc_name}, try checking if it exists, or is attached"
        )
        exit(1)
    with client.ApiClient() as api_client:
        if pod_info["used_by"] != "<none>":
            api_instance = client.CoreV1Api(api_client)
            try:
                # TODO This may come back to bite me in the ass. If it grabs the wrong object, it's because we
                # assume there's only one owner reference
                if namespace == "":
                    if "namespace" in pod_info.keys():
                        pod_namespace = pod_info["namespace"]

                api_response = api_instance.read_namespaced_pod(
                    pod_info["used_by"], pod_namespace
                )
                if api_response.metadata.owner_references is not None:
                    return (
                        api_response.metadata.owner_references[0].name,
                        pod_namespace,
                        api_response.metadata.owner_references[0].kind,
                    )
                else:
                    return ("", "", "")
            except client.exceptions.ApiException:
                logging.error(
                    f"Could not find pod {pod_info['used_by']} for pvc {pvc_name}"
                )
                exit(1)
        else:
            logging.info(
                f"Could not find a pod for pvc {pvc_name}, going to assume its currently unattached"
            )
            return ("", "", "")


def _get_information_from_description(description: str) -> dict:
    """
    Given the printout of a kubernetes describe command for various Kubernetes objects, extract
    key information and return it as a dictionary
    Args:
        description (str): The string containing the output of a kubernetes describe command

    Returns:
        dict[str, str]: A dictionary that may contain the following keys:
        used_by: What pod a PVC is used by
        namespace: The namespace of the object
        controlled_by: The Deployment or other controller that manages a ReplicaSet
    """
    return_dict = {}
    # TODO error handle here if needed
    for line in description.split("\n"):
        if "Used By:" in line:
            splitted = line.split(" ")
            return_dict["used_by"] = splitted[len(splitted) - 1]
        elif "Namespace:" in line:
            splitted = line.split(" ")
            return_dict["namespace"] = splitted[len(splitted) - 1]
        elif "Controlled By:" in line:
            splitted = line.split(" ")
            raw_value = splitted[len(splitted) - 1]
            return_dict["controlled_by"] = raw_value.split("/")[1]
    return return_dict


def _get_pvc_description(pvc_name: str, namespace: str) -> str:
    if not namespace:
        result = subprocess.run(
            ["kubectl", "describe", "pvc", pvc_name], capture_output=True, text=True
        )
    else:
        result = subprocess.run(
            ["kubectl", "describe", "pvc", pvc_name, "-n", namespace],
            capture_output=True,
            text=True,
        )

    return result.stdout


def _launch_job(job_dict):
    job_name = job_dict["metadata"]["name"]

    if "namespace" in job_dict["metadata"]:
        namespace = job_dict["metadata"]["namespace"]
    else:
        _, active_context = k8s_config.list_kube_config_contexts()
        if "namespace" in active_context["context"]:
            namespace = active_context["context"]["namespace"]
        else:
            namespace = "default"
        job_dict["metadata"]["namespace"] = namespace

    _clear_job_name(job_name, namespace)

    k8s_client = client.ApiClient()
    utils.create_from_dict(k8s_client, job_dict)


def _modify_controller_replicas(
    controller_name: str, namespace: str, controller_type: str, num_replicas: int
) -> int:
    """
    Modify the number of replicas a controller is requesting
    Args:
        controller_name (str): The name of the controller
        namespace (str): The namespace the controller lives in
        controller_type (str): The kind (i.e. ReplicaSet, DaemonSet, etc) of the controller
        num_replicas (int): The number of replicas we want to set for the controller
    """
    with client.ApiClient() as api_client:
        api_instance = client.AppsV1Api(api_client)
        try:
            if controller_type == "ReplicaSet":
                result = subprocess.run(
                    [
                        "kubectl",
                        "describe",
                        "replicaset",
                        controller_name,
                        "-n",
                        namespace,
                    ],
                    capture_output=True,
                    text=True,
                )
                info = _get_information_from_description(result.stdout)

                previous_replicas: int = api_instance.read_namespaced_deployment(
                    info["controlled_by"], namespace
                ).spec.replicas

                patch_body = (
                    '[{"op": "replace", "path": "/spec/replicas", "value":'
                    + str(num_replicas)
                    + "}]"
                )
                api_response = api_instance.patch_namespaced_deployment(
                    info["controlled_by"], namespace, json.loads(patch_body)
                )

                return previous_replicas
        except client.exceptions.ApiException as e:
            logging.error(f"Failed to scale down {controller_name}")
            logging.debug(e)
            return -1
    return -1


def _render_job(job_name: str, args: dict[str, str]) -> str:
    """
    Given the name of a job, that is the same as a directory in the JOBS_DIR directory,
    render the template with the arguments provided, and return it
    Args:
        job_name (str): The name a job in the JOBS_DIR directory
        args (dict[str, str]): A dictionary representing arguments. Each key should correspond to a
        variable in a job template, with each value representing what should be filled in
    Returns:
        str: The contents of the template file, rendered with the values of args
    """
    jobs_dir = config.get_full_atils_dir("JOBS_DIR")
    if os.path.exists(os.path.join(jobs_dir, job_name, "job.yaml")):
        rendered_job = template_utils.template_external_file(
            os.path.join(jobs_dir, job_name, "job.yaml"), args
        )
        return yaml.safe_load(rendered_job)

    else:
        logging.error(f'Job "{job_name}" was not found')
        exit(1)


def _scale_down_controller_from_pvc(pvc_name: str) -> int:
    """
    Given the name of a pvc, identify any pod it is attached to, and from that pod, identify
    a corresponding controller. After that, scale down the controller, and return how many
    replicas it had previously. If no controller was found, return -1

    Args:
        pvc_name (str): The name of the pvc whose pod we want to disable

    Returns:
        int: -1 if no controller was found, and the number of replicas has the controller
        set as desired if a controller was found
    """
    (
        controller_name,
        controller_namespace,
        controller_kind,
    ) = _get_controller_from_pvc(pvc_name)

    # We'll assume by default that previous replicas was 1, unless we find
    # something better
    previous_replicas = -1
    if controller_name is not None:
        previous_replicas = _modify_controller_replicas(
            controller_name, controller_namespace, controller_kind, 0
        )

    return previous_replicas



import argparse
import asyncio
import logging
import os
import subprocess
import sys
import time

import watchfiles

from atils.common import config
from atils.common.settings import settings

logging.basicConfig(level=config.get_logging_level())  # type: ignore
logging.getLogger("asyncio").setLevel(logging.WARNING)
watchfiles.main.logger.setLevel(logging.WARNING)


def main(args: str):
    parser: argparse.ArgumentParser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        help="Select a subcommand", dest="subparser_name"
    )

    auto_deploy_parser = subparsers.add_parser(
        "auto-deploy", help="Watch a helm chart for changes, and deploy automatically"
    )

    auto_deploy_parser.add_argument(
        "--chart-name",
        "-cn",
        type=str,
        help="The name of the helm chart to watch",
        required=True,
    )

    if len(args) == 0:
        parser.print_help(sys.stderr)
        sys.exit(1)

    arguments: argparse.Namespace = parser.parse_args(args)

    # Let's explain this a little bit. We run the auto_deploy_helm_chart async, so that when we get a
    # KeyboardInterrupt (i.e. Ctrl+C ), we can uninstall the chart automatically.
    if arguments.subparser_name == "auto-deploy":
        try:
            asyncio.run(auto_deploy_helm_chart(arguments.chart_name))
        except KeyboardInterrupt:
            subprocess.run(["helm", "uninstall", arguments.chart_name])


async def auto_deploy_helm_chart(chart_name: str) -> None:
    """
    Automatically deploy a helm chart in the HELM_CHARTS_DIR directory. After the initial deployment,
    it then reinstalls the chart whenever an update is made to the chart, or to the values file located at
    ~/.atils/helm-values/<chart-name>.yaml

    Args:
        chart_name (str): The name of the helm chart to deploy.
    """
    helm_chart_dir = os.path.join(
        settings.install_dir, settings.helm_charts_dir, chart_name
    )

    # Check if there is a directory at the given path, that has a Chart.yaml file in it
    if os.path.isdir(helm_chart_dir) and os.path.isfile(
        os.path.join(helm_chart_dir, "Chart.yaml")
    ):
        print(f"Watching and auto-installing {chart_name}...")
        _install_helm_chart(chart_name, helm_chart_dir)
        # This is where we watch for changes to the chart and values file. Those specific functions then
        # handle calling the install function
        await asyncio.gather(
            _watch_chart_path(chart_name), _watch_values_path(chart_name)
        )

    else:
        logging.error(f"No chart found at the given path: {helm_chart_dir}")
        exit(1)


def _install_helm_chart(chart_name: str, helm_chart_dir: str) -> None:
    """
    Automatically installs a helm chart located at helm_chart_dir. If there is a file at
    ~/.atils/helm-values/<chart_name>.yaml, this functions uses that as the values.yaml file

    Args:
        chart_name (str): The name of the helm chart to install, which is used as the release name
        helm_chart_dir (str): The path to the helm chart to install
    """
    values_file = os.path.join(
        settings.config_directory, "helm-values", f"{chart_name}.yaml"
    )
    if os.path.isfile(values_file):
        subprocess.run(
            [
                "helm",
                "upgrade",
                "--install",
                chart_name,
                helm_chart_dir,
                "-f",
                values_file,
            ]
        )
    else:
        subprocess.run(["helm", "upgrade", "--install", chart_name, helm_chart_dir])


async def _watch_chart_path(chart_name: str) -> None:
    """
    Watches a local helm chart for changes, and calls _install_helm_chart whenever an update is detected

    Args:
        chart_name (str): The name of the chart. This will be the name of a directory in the HELM_CHARTS_DIR
    """
    helm_chart_dir = os.path.join(
        settings.install_dir, settings.helm_charts_dir, chart_name
    )

    async for change in watchfiles.awatch(helm_chart_dir):
        _install_helm_chart(chart_name, helm_chart_dir)


async def _watch_values_path(chart_name: str) -> None:
    """
    Watches the values file in ~/.atils/helm-values/<chart_name>.yaml for changes, and calls _install_helm_chart
    whenever an update is detected

    Args:
        chart_name (str): The name of the chart. This will be the name of a directory in the HELM_CHARTS_DIR
    """
    helm_chart_dir = os.path.join(
        settings.install_dir, settings.helm_charts_dir, chart_name
    )

    values_file = os.path.join(
        settings.config_directory, "helm-values", f"{chart_name}.yaml"
    )
    # TODO we may want to clean this up, I don't think we need to loop in this function, when we're calling it in a loop
    while True:
        if os.path.isfile(values_file):
            async for change in watchfiles.awatch(values_file):
                _install_helm_chart(chart_name, helm_chart_dir)
        else:
            time.sleep(5)
